/*
 * ### Wochenschaltuhr ###
 * to use with AVR-NETIO hardware by <pollin.de>
 * and DS3231 RTC, I2C connected at Pin17 & Pin18  
 * 
 *      -----------------
 *
 * it will respond to pings
 *
 * it was based upon uIP by Adam Dunkels <adam@sics.se>
 * ported to Arduino by Adam Nielsen <malvineous@shikadi.net>
 * adapted to Enc28J60 by Norbert Truchsess <norbert.truchsess@t-online.de>
 * Soft-I2C-library found on <github.com/felias-fogg/SoftI2CMaster>
 */

#define MACADDRESS 0x00,0x22,0x44,0x66,0x88,0xAA   // define MAC address
#define UARTBAUD 9600
#define PULSETIME 1000          // output impulse duration
#define eeAUSG 2047       // eeprom location output state
#define eeAUTO 2046       // eeprom location automatic state
#define eePortH 2045      // eeprom location port high
#define eePortL 2044      // eeprom location port low
#define eeIP1 2043        // eeprom location ip-adress part 1 (high) 
#define eeIP2 2042        // eeprom location ip-adress part 2
#define eeIP3 2041        // eeprom location ip-adress part 3
#define eeIP4 2040        // eeprom location ip-adress part 4 (low) 

#include <UIPEthernet.h>
#include <UIPServer.h>
#include <UIPClient.h>
#include <EEPROM.h>
#include <avr/wdt.h>

#define LOAD_PIN 14       // button load - Port D6 (negative pulse)
#define RCK_PIN 12        // led enable - Port D4
#define CLK_PIN 13        // led + button clock - Port D5
#define DAT_OUT_PIN 15    // led out Port D7
#define DAT_IN_PIN 0      // buttons in Port B0
#define RST_PIN 29        // factory reset pin - clr the eeprom  Port A5
#define SCL_PIN 3         // RTC  Port D3  
#define SCL_PORT PORTD    // RTC
#define SDA_PIN 2         // RTC  Port D2
#define SDA_PORT PORTD    // RTC
#define RTC_I2C_ADDRESS 0x68 // I2C Adresse des RTC DS3231
#include <SoftI2CMaster.h>

byte wochentag, stunde, minute, sekunde, ausgang, automatik, impuls=0, prevmin;
word weektick;
unsigned long prevMillis = 0, pre_ms = 0;      // timecounter
EthernetServer server = EthernetServer(EEPROM[eePortH] * 256 + EEPROM[eePortL]);

void setup(){
  for (byte ii = 16; ii < 24; ii++){
    pinMode(ii, OUTPUT);                     //  init the output pins (PortC 0...7)
  }
  pinMode(RST_PIN, INPUT_PULLUP);            //  init the factory reset pin
  pinMode(RCK_PIN, OUTPUT);                  //  init the led enable pin
  pinMode(LOAD_PIN, OUTPUT);                 //  init the button load pin
  digitalWrite(LOAD_PIN, HIGH);              //  set the button load pin high
  pinMode(CLK_PIN, OUTPUT);                  //  init the clock pin
  pinMode(DAT_OUT_PIN, OUTPUT);              //  init the button data-out pin
  pinMode(DAT_IN_PIN, INPUT);                //  init the led data-in pin
  Serial.begin(UARTBAUD);
  if(digitalRead(RST_PIN) == LOW){           // if button is pressed, all events a erased
    for (word ii = 0; ii < 1750; ii += 7){   // setzen der Index-Zellen ... 
      if (EEPROM[ii] != 0xFF){
        EEPROM[ii] = 0xFF;                   // ... auf 0xFF
      } 
    }
    EEPROM.update(eeAUSG, 0x00);
    EEPROM.update(eeAUTO, 0xFF);
    EEPROM.update(eePortH, 26);
    EEPROM.update(eePortL, 176);
    EEPROM.update(eeIP1, 192);
    EEPROM.update(eeIP2, 168);
    EEPROM.update(eeIP3, 1);
    EEPROM.update(eeIP4, 1);
    Serial.println("eeprom delete");
  }
  uint8_t mac[6] = {MACADDRESS};
  IPAddress myIP(EEPROM[eeIP1], EEPROM[eeIP2], EEPROM[eeIP3], EEPROM[eeIP4]);
  Ethernet.begin(mac,myIP);
  server.begin();
  Serial.print("start server @ ");
  Serial.print(Ethernet.localIP());
  Serial.print(":"); Serial.print(EEPROM[eePortH] * 256 + EEPROM[eePortL]);
  Serial.print("\n");
  ausgang = EEPROM[eeAUSG];
  automatik = EEPROM[eeAUTO];
  setOutputs(false);
  setLEDs();
}


byte getKeystroke(){
  static byte keystore;
  byte temp = 0;
  digitalWrite(LOAD_PIN, LOW);      // catch the button state
  digitalWrite(LOAD_PIN, HIGH);
  for (byte ii = 0; ii < 8; ii++){
    if (digitalRead(DAT_IN_PIN)){
      bitSet(temp, ii);
    }
    digitalWrite(CLK_PIN, HIGH);    // clock in the next state
    digitalWrite(CLK_PIN, LOW);
  }
  byte keys = (temp << 4) & B11110000;      // move the 4 low bits to high
  keys |= ((temp >> 4) & B00001111);       // move the 4 high bits to low 
  if (keys == 0){
    keystore = 0;
    return 0;
  }
  else{
    if (keys == keystore){
      return 0;
    }
    else{
      delay(10);
      keystore = keys;
      return keys;          // return the key one time per stroke
    }
  }
}


void setLEDs(){
  for (byte ii = 0; ii < 8; ii++){
    if (automatik & (128 >> ii)){
      digitalWrite(DAT_OUT_PIN, LOW);
    }
    else{  
      digitalWrite(DAT_OUT_PIN, HIGH);
    }
    digitalWrite(CLK_PIN, HIGH);
    digitalWrite(CLK_PIN, LOW);
  }
  digitalWrite(RCK_PIN, HIGH);
  digitalWrite(RCK_PIN, LOW);
  EEPROM.update(eeAUTO, automatik);
}


// Hier werden die Taster ausgewertet.
// Die 8 Taster sind den 8 Bit in den Variablen 'ausgang' und 'automatik' zugeordnet.
// Jeder Taster schaltet die zwei zugeordneten Bits, d.h. es gibt vier Schaltzustände.
void serveKeyboard(){
  byte keys = getKeystroke();
  if(keys > 0){
    if (ausgang & keys){
      ausgang &= ~keys;         // clear the bit
      if (automatik & keys){
        automatik &= ~keys;     // clear the bit  
      }
      else{
        automatik |= keys;
      }
    }
    else{
      ausgang |= keys;  
    }
    setOutputs(false);
    setLEDs();
  }  
}


// konvertiert 2-x-4-Bit-BCD-Zahlen in Dezimalzahlen 
byte bcdToDec(byte val){
  return ((val/16*10) + (val%16));
}


// konvertiert Dezimalzahlen in 2-x-4-Bit-BCD-Zahlen
byte decToBcd(byte val){
  return ( (val/10*16) + (val%10) );
}


//Liest die aktuelle Zeit und den Wochentag aus der RTC
void getRtcTime(){
  i2c_start((RTC_I2C_ADDRESS<<1)|I2C_WRITE);
  i2c_write(0); // setzt den RegisterPointer im RTC auf Sekunde
  i2c_rep_start((RTC_I2C_ADDRESS<<1)|I2C_READ); // restart for reading
  sekunde = bcdToDec(i2c_read(true) & 0x7f);
  i2c_write(1); // setzt den RegisterPointer im RTC auf Minute
  i2c_rep_start((RTC_I2C_ADDRESS<<1)|I2C_READ); // restart for reading 
  minute = bcdToDec(i2c_read(true) & 0x7f); 
  i2c_write(2); // setzt den RegisterPointer im RTC auf Stunde
  i2c_rep_start((RTC_I2C_ADDRESS<<1)|I2C_READ); // restart for reading
  stunde = bcdToDec(i2c_read(true) & 0x3f); 
  i2c_write(3); // setzt den RegisterPointer im RTC auf Wochentag
  i2c_rep_start((RTC_I2C_ADDRESS<<1)|I2C_READ); // restart for reading
  wochentag = bcdToDec(i2c_read(true) & 0x07); 
  i2c_stop(); // send stop condition
}


// setzt die Uhrzeit und den Wochentag in der RTC auf neuen Wert
void setRTCTime(uint8_t* msg){
  i2c_start((RTC_I2C_ADDRESS<<1)|I2C_WRITE);
  i2c_write(0); // setzt den RegisterPointer im RTC auf Sekunde
  i2c_write(decToBcd(msg[11])); // secund - der Pointer incrementiert nach jedem Zugriff automatisch
  i2c_write(decToBcd(msg[12])); // minute
  i2c_write(decToBcd(msg[13])); // hour                                 
  i2c_write(msg[14]);           // day of week
  i2c_stop();  
}


// setzt die IP-Adresse und den Port auf neuen Wert
void setIP(uint8_t* msg){
  EEPROM.update(eePortH, msg[14]);
  EEPROM.update(eePortL, msg[13]);
  EEPROM.update(eeIP1, msg[12]);
  EEPROM.update(eeIP2, msg[11]);
  EEPROM.update(eeIP3, msg[10]);
  EEPROM.update(eeIP4, msg[9]);
  Serial.println("set ip + port, restart");
  // start watchdog with the provided prescaller
  wdt_enable(WDTO_60MS);
  // wait for the prescaller time to expire
  // without sending the reset signal by using
  // the wdt_reset() method
  while(1) {}
}

// aut = 0 : Setzt die Ausgänge 
// aut = 1 : Setzt die Ausgänge bei denen Auomatik eingeschaltet ist 
// PortC (Pins = 16, 17, 18, 19, 20, 21, 22, 23) gemäß 'ausgang' einstellen
void setOutputs(bool aut){
  byte temp = ausgang;
  if (aut){
    temp = ausgang & automatik;
  }
  for (int ii = 0; ii < 8; ii++){
    if (temp & (1<<ii)){
      digitalWrite(ii + 16, HIGH);
    }  
    else{
      digitalWrite(ii + 16, LOW);
    }
  }
  EEPROM.update(eeAUSG, temp);
}


// Ausgabe der Systemzeit über die RS232
void prntTime(){
  Serial.print(wochentag);
  Serial.print(" ");
  if (stunde < 10) Serial.print("0"); 
  Serial.print(stunde);
  Serial.print(":");
  if (minute < 10) Serial.print("0");
  Serial.print(minute);
  Serial.print(":");
  if (sekunde < 10) Serial.print("0");
  Serial.print(sekunde);
}


boolean verifyCheckSum(uint8_t* msg){
  byte tmp = 0;
  for (int ii = 0; ii < 15; ii++){
    tmp ^= msg[ii];
  }
  if (tmp == msg[15]){ return true; }
  else{ return false; }
}


// *Index, *Zeit-L, *Zeit-H, *Ausgänge-Ein, *Ausgänge-Aus, *Ausgänge-Impuls, *Deaktiv-Löschen
// wird jede Minute aufgerufen und durchsucht das EEPROM nach aktuell auszuführenden Aufgaben
boolean isTimeToSwitch(){
  boolean result = false;
  word zeit;
  for (word ii = 0; ii < 1750; ii += 7){   // testen der Index-Zellen ... 
    if (EEPROM[ii] < 0xFF){                // ... auf gültige Einträge
      zeit = EEPROM[ii+1];                 // read L-Byte
      zeit = zeit + (EEPROM[ii+2] * 256);  // add H-Byte
      if(zeit == weektick){
        if(bitRead(EEPROM[ii+6], 0) == 1){ // wenn das Aktiviert-Bit gesetzt ist  
          ausgang |= EEPROM[ii+3];         // einzuschaltende Ausgänge übertragen
          ausgang &= (~EEPROM[ii+4]);      // auszuschaltende Ausgänge übertragen
          impuls = EEPROM[ii+5];           // Wert zum Setzen und Rücksetzen speichern
          ausgang ^= impuls;               // ExOr des entsprechenden Ausgangs
        }
        if(bitRead(EEPROM[ii+6], 1) == 1){ // Löschenflag auswerten
          EEPROM[ii] = 0xFF;               // Index des Eintrags als gelöscht markieren
        }
        result = true;   
      }
    }
  }
  return result;
}


// neuen Event in das EEPROM eintragen
// $55, $AC, $00, $00, *Index, *Zeit-L, *Zeit-H, *Ausgänge-Ein, *Ausgänge-Aus, *Ausgänge-Impuls, *Flags
void setNewEvent(uint8_t* msg){
  boolean stored = false;
  for (int ii = 0; ii < 1750; ii += 7){ // testen der Index-Zellen ... 
    if (EEPROM[ii] == 0xFF){            // ... auf freien Platz
      EEPROM[ii] = msg[4];              // Index
      EEPROM.update(ii+1, msg[5]);      // Zeit-L
      EEPROM.update(ii+2, msg[6]);      // Zeit-H
      EEPROM.update(ii+3, msg[7]);      // Ausgänge-Ein
      EEPROM.update(ii+4, msg[8]);      // Ausgänge-Aus
      EEPROM.update(ii+5, msg[9]);      // Ausgänge-Impuls
      EEPROM.update(ii+6, msg[10]);     // Deaktiv-Löschen Flags
      stored = true;
      break;                            // done, exit the for loop
    }
  }
  if(!stored) Serial.print("error - eeprom overflow");
}


// Event löschen
// $55, $AC, $00, $00, *Index, *Zeit-L, *Zeit-H, *Ausgänge-Ein, *Ausgänge-Aus, *Ausgänge-Impuls, *Deaktiv-Löschen
void deleteEvent(byte index){
  for (int ii = 0; ii < 1750; ii += 7){    // alle Index-Zellen im EEPROM ... 
    if (EEPROM[ii] == index){              // ... auf den bestimmten Index-Eintrag testen, ...
      EEPROM[ii] = 0xFF;              // ... dann mit $FF als gelöscht markieren
    }
  }
}


// $55, $AA, Ausgangsstatus, Automatikstatus, Aufgaben, Reserve, *Index, *Zeit-L, *Zeit-H, *Ausgänge-Ein, *Ausgänge-Aus, *Ausgänge-Impuls, *Flags, Checksumme
void sendGlobalFrame(EthernetClient client){
  byte msg[1757];
  word cnt = 6;
  byte jj = 0;
  for (word ii = 0; ii < 1757; ii++){
    msg[ii] = 0; 
  }
  msg[0] = 0x55;
  msg[1] = 0xAA;
  msg[2] = ausgang;
  msg[3] = automatik;
  // look for stored events in eeprom
  for (byte ii = 0; ii < 250; ii += 7){
    if (EEPROM[ii] != 0xFF){
      msg[cnt] = EEPROM[ii];
      msg[cnt+1] = EEPROM[ii+1];
      msg[cnt+2] = EEPROM[ii+2];
      msg[cnt+3] = EEPROM[ii+3];
      msg[cnt+4] = EEPROM[ii+4];
      msg[cnt+5] = EEPROM[ii+5];
      msg[cnt+6] = EEPROM[ii+6];
      cnt += 7;
      jj ++;
    }
  }
  // number of events inside this frame
  msg[4] = jj;
  // checksum calculation
  byte chksm = 0;
  for (byte ii = 0; ii <= cnt; ii ++){
    chksm ^= msg[ii];
  }
  msg[cnt+1] = chksm;
  // send the frame
  client.write(msg,cnt+3);  
}


void working(){
  unsigned long currentMillis = millis();   // get the arduino timecounter 
  if (currentMillis - prevMillis >= 1000){  // interval (milliseconds)
    prevMillis = currentMillis;             // save the last timecounter
    getRtcTime();                           // every second
    if (prevmin != minute){                 // do it each minute 
      prevmin = minute;    
      weektick = ((wochentag-1) * 1440) + (stunde * 60) + minute; 
      if (isTimeToSwitch()){                // compare the whole event list
        prntTime(); Serial.println(" do event");
        setOutputs(true);
      }
      if (impuls > 0){
        delay(PULSETIME);                   // quick and dirty :-)
        ausgang ^= impuls;
        setOutputs(true);
        impuls = 0;
      }
    }
  }    
}


void loop(){
  size_t size;
  if (EthernetClient client = server.available()){
    Serial.println("ui connected");
    while (client.connected()) {  // loop with connected user interface
      while((size = client.available()) > 0){
        uint8_t* msg = (uint8_t*)malloc(size);
        size = client.read(msg,size);
        byte cnt = 0;
        while (size > 15){
          if ((msg[cnt] == 0x55) && verifyCheckSum(msg+cnt)){ // wenn 1. Byte = $55 und Checksumme stimmt
            switch (msg[cnt+1]){
              case 0xA8:            // time set
                setRTCTime(msg+cnt);
                Serial.println("adjusting the rtc");
                break;
              case 0xA9:            // ip-adress + port set
                setIP(msg+cnt);     // reset the mcu
                break;
              case 0xAB:            // send a global frame
                sendGlobalFrame(client);
                Serial.println("send a global frame");
                break;
              case 0xAC:            // new event
                setNewEvent(msg+cnt);
                Serial.println("set a new event");
                break;
              case 0xAD:            // switch the output and automatic status now
                ausgang = msg[cnt+2];
                automatik = msg[cnt+3];
                setOutputs(false);
                setLEDs();
                Serial.println("switch the output");
                break;
              case 0xAE:            // delet event
                deleteEvent(msg[cnt+4]);
                Serial.println("delete a event");
                break;
              case 0xAF:            // keep alive
                // do nothing
                prntTime(); Serial.println(" ui still connected");
                break;
            }
          }  
          cnt = cnt + 16; 
          size = size - 16; 
        }
        size = 0;  
        free(msg);
        pre_ms = millis();                // renew the value for connection timeout
      }
      if (millis() - pre_ms >= 11000){    // interval 11 sec.
        break;                            // connection timeout
      }
      working();
      serveKeyboard();
    }
    client.stop();
    Serial.println("ui disconnected");
  }
  working();
  serveKeyboard();
}
